import logging
import shutil
from copy import deepcopy
from textwrap import indent
from typing import Generator, Iterable, List, NamedTuple, Optional, Sequence, Set, Tuple

from colorama import Fore, Style  # type: ignore

from .actions.action import Action
from .compat import Path
from .config import Rule
from .filters.filter import Filter
from .utils import DotDict, splitglob

logger = logging.getLogger(__name__)
SYSTEM_FILES = ("thumbs.db", "desktop.ini", ".DS_Store")

Job = NamedTuple(
    "Job",
    [
        ("folderstr", str),
        ("basedir", Path),
        ("path", Path),
        ("filters", Sequence[Filter]),
        ("actions", Sequence[Action]),
    ],
)
Job.__doc__ = """
    :param str folderstr: the original folder definition specified in the config
    :param Path basedir:  the job's base folder
    :param Path path:     the path of the file to handle
    :param list filters:  the filters that apply to the path
    :param list actions:  the actions which should be executed
"""


class OutputHelper:
    """
    class to track the current folder / file and print only changes.
    This is needed because we only want to output the current folder and file if the
    filter or action prints something.
    """

    def __init__(self) -> None:
        self.not_found = set()  # type: Set[str]
        self.curr_folder = None  # type: Optional[Path]
        self.curr_path = None  # type: Optional[Path]
        self.prev_folder = None  # type: Optional[Path]
        self.prev_path = None  # type: Optional[Path]

    def set_location(self, folder: Path, path: Path) -> None:
        self.curr_folder = folder
        self.curr_path = path

    def pre_print(self) -> None:
        """
        pre-print hook that is called everytime the moment before a filter or action is
        about to print something to the cli
        """
        if self.curr_folder != self.prev_folder:
            if self.prev_folder is not None:
                print()  # ensure newline between folders
            print("Folder %s%s:" % (Style.BRIGHT, self.curr_folder))
            self.prev_folder = self.curr_folder

        if self.curr_path != self.prev_path:
            print(indent("File %s%s:" % (Style.BRIGHT, self.curr_path), " " * 2))
            self.prev_path = self.curr_path

    def print_path_not_found(self, folderstr: str) -> None:
        if folderstr not in self.not_found:
            self.not_found.add(folderstr)
            msg = "Path not found: {}".format(folderstr)
            print(Fore.YELLOW + Style.BRIGHT + msg)
            logger.warning(msg)


output_helper = OutputHelper()


def execute_rules(rules: Iterable[Rule], simulate: bool) -> None:
    cols, _ = shutil.get_terminal_size(fallback=(79, 20))
    simulation_msg = Fore.GREEN + Style.BRIGHT + " SIMULATION ".center(cols, "~")

    jobs = create_jobs(rules=rules)

    if simulate:
        print(simulation_msg)

    failed, succeded = run_jobs(jobs=jobs, simulate=simulate)
    if succeded == failed == 0:
        msg = "Nothing to do."
        logger.info(msg)
        print(msg)

    if simulate:
        print(simulation_msg)


def create_jobs(rules: Iterable[Rule]) -> Generator[Job, None, None]:
    """ creates `Job` data structures for every path handled in each rule """
    for rule in rules:
        for folderstr, basedir, path in all_files_for_rule(rule):
            yield Job(
                folderstr=folderstr,
                basedir=basedir,
                path=path,
                filters=rule.filters,
                actions=rule.actions,
            )


def all_files_for_rule(rule: Rule) -> Generator[Tuple[str, Path, Path], None, None]:
    files = dict()
    for folderstr in rule.folders:
        folderstr = folderstr.strip()

        # check whether the file / folder is prefixed with `!` to be excluded
        exclude_flag = folderstr.startswith("!")

        # assemble glob expression
        basedir, globstr = splitglob(folderstr.lstrip("!").strip())
        if basedir.is_dir():
            if not globstr:
                globstr = "**/*" if rule.subfolders else "*"
        elif basedir.is_file():
            # this allows specifying single files
            globstr = basedir.name
            basedir = basedir.parent
        else:
            output_helper.print_path_not_found(folderstr)
            continue

        # iterate files in basedir and add to / remove from result dict
        for path in basedir.glob(globstr):
            if path.is_file() and (rule.system_files or path.name not in SYSTEM_FILES):
                if not exclude_flag:
                    files[path] = (folderstr, basedir)
                elif path in files:
                    del files[path]

    for path, (folderstr, basedir) in files.items():
        yield (folderstr, basedir, path)


def run_jobs(jobs: Iterable[Job], simulate: bool) -> List[int]:
    """ :returns: The number of successfully handled files """
    count = [0, 0]
    Action.pre_print_hook = output_helper.pre_print
    Filter.pre_print_hook = output_helper.pre_print

    for job in sorted(jobs, key=lambda x: (x.folderstr, x.basedir, x.path)):
        args = DotDict(path=job.path, basedir=job.basedir, simulate=simulate)
        # the relative path should be kept even if the path changes.
        args["relative_path"] = args["path"].relative_to(args["basedir"])

        output_helper.set_location(job.basedir, args.relative_path)
        match = filter_pipeline(filters=job.filters, args=args)
        if match:
            success = action_pipeline(actions=job.actions, args=args)
            count[success] += 1
    return count


def filter_pipeline(filters: Iterable[Filter], args: DotDict) -> bool:
    """
    run the filter pipeline.
    Returns True on a match, False otherwise and updates `args` in the process.
    """
    for filter_ in filters:
        try:
            result = filter_.pipeline(deepcopy(args))
            if isinstance(result, dict):
                args.update(result)
            elif not result:
                # filters might return a simple True / False.
                # Exit early if a filter does not match.
                return False
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            filter_.print(Fore.RED + Style.BRIGHT + "ERROR! %s" % e)
            return False
    return True


def action_pipeline(actions: Iterable[Action], args: DotDict) -> bool:
    for action in actions:
        try:
            updates = action.pipeline(deepcopy(args))
            # jobs may return a dict with updates that should be merged into args
            if updates is not None:
                args.update(updates)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            action.print(Fore.RED + Style.BRIGHT + "ERROR! %s" % e)
            return False
    return True



import re
from typing import Any, Dict, Mapping, Optional

import textract  # type: ignore
from organize.compat import Path

from .filter import Filter


# not supported: .gif, .jpg, .mp3, .ogg, .png, .tiff, .wav
SUPPORTED_EXTENSIONS = (
    ".csv .doc .docx .eml .epub .json .html .msg .odt .pdf .pptx .ps .rtf .txt .xlsx .xls"
).split()


class FileContent(Filter):

    r"""
    Matches file content with the given regular expression

    :param str expr:
        The regular expression to be matched.

    Any named groups in your regular expression will be returned like this:

    :returns:
        - ``{filecontent.yourgroupname}`` -- The text matched with the named group
          ``(?P<yourgroupname>)``

    Examples:

        - Match an invoice with a regular expression and sort by customer:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - filecontent: 'Invoice.*Customer (?P<customer>\w+)'
                actions:
                  - move: '~/Documents/Invoices/{filecontent.customer}/'
    """

    def __init__(self, expr) -> None:
        self.expr = re.compile(expr, re.MULTILINE | re.DOTALL)

    def matches(self, path: Path) -> Any:
        if path.suffix.lower() not in SUPPORTED_EXTENSIONS:
            return
        try:
            content = textract.process(str(path), errors="ignore")
            return self.expr.search(content.decode("utf-8", errors="ignore"))
        except textract.exceptions.CommandLineError:
            pass

    def pipeline(self, args: Mapping) -> Optional[Dict[str, Dict]]:
        match = self.matches(args["path"])
        if match:
            result = match.groupdict()
            return {"filecontent": result}
        return None



import operator
import re
from typing import Callable, Dict, Optional, Sequence, Set, Tuple

from organize.utils import DotDict, flattened_string_list, fullpath

from .filter import Filter

OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "=": operator.eq,
    "": operator.eq,
    ">=": operator.ge,
    ">": operator.gt,
}
SIZE_REGEX = re.compile(
    r"^(?P<op>[<>=]*)(?P<num>(\d*\.)?\d+)(?P<unit>[kmgtpezy]?i?)b?$"
)


def create_constrains(inp: str) -> Set[Tuple[Callable[[int, int], bool], int]]:
    """
    Given an input string it returns a list of tuples (comparison operator,
    number of bytes).

    Accepted formats are: '30k', '>= 5 TiB, <10tb', '< 60 tb', ...
    Calculation is in bytes, even if the 'b' is lowercase. If an 'i' is present
    we calculate base 1024.
    """
    result = set()  # type: Set[Tuple[Callable[[int, int], bool], int]]
    parts = inp.replace(" ", "").lower().split(",")
    for part in parts:
        try:
            reg_match = SIZE_REGEX.match(part)
            if reg_match:
                match = reg_match.groupdict()
                op = OPERATORS[match["op"]]
                num = float(match["num"]) if "." in match["num"] else int(match["num"])
                unit = match["unit"]
                base = 1024 if unit.endswith("i") else 1000
                exp = "kmgtpezy".index(unit[0]) + 1 if unit else 0
                numbytes = num * base ** exp
                result.add((op, numbytes))
        except (AttributeError, KeyError, IndexError, ValueError, TypeError) as e:
            raise ValueError("Invalid size format: %s" % part) from e
    return result


def satisfies_constrains(size, constrains):
    return all(op(size, p_size) for op, p_size in constrains)


class FileSize(Filter):
    """
    Matches files by file size

    :param str conditions:

    Accepts file size conditions, e.g: ``'>= 500 MB'``, ``'< 20k'``, ``'>0'``,
    ``'= 10 KiB'``.

    It is possible to define both lower and upper conditions like this:
    ``'>20k, < 1 TB'``, ``'>= 20 Mb, <25 Mb'``. The filter will match if all given
    conditions are satisfied.

    - Accepts all units from KB to YB.
    - If no unit is given, kilobytes are assumend.
    - If binary prefix is given (KiB, GiB) the size is calculated using base 1024.

    :returns:
        - ``{filesize.bytes}`` -- File size in bytes

    Examples:
        - Trash big downloads:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Downloads'
                filters:
                  - filesize: '> 0.5 GB'
                actions:
                  - trash

        - Move all JPEGS bigger > 1MB and <10 MB. Search all subfolders and keep the´
          original relative path.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Pictures'
                subfolders: true
                filters:
                  - extension:
                      - jpg
                      - jpeg
                  - filesize: '>1mb, <10mb'
                actions:
                  - move: '~/Pictures/sorted/{relative_path}/'

    """

    def __init__(self, *conditions: Sequence[str]) -> None:
        self.conditions = ", ".join(flattened_string_list(list(conditions)))
        self.constrains = create_constrains(self.conditions)
        if not self.constrains:
            raise ValueError("No size(s) given!")

    def matches(self, filesize: int) -> bool:
        return all(op(filesize, c_size) for op, c_size in self.constrains)

    def pipeline(self, args: DotDict) -> Optional[Dict[str, Dict[str, int]]]:
        file_size = fullpath(args.path).stat().st_size
        if self.matches(file_size):
            return {"filesize": {"bytes": file_size}}
        return None

    def __str__(self) -> str:
        return "FileSize({})".format(" ".join(self.conditions))



import sys
from typing import Dict, Optional, SupportsFloat

import pendulum  # type: ignore
from organize.compat import Path
from organize.utils import DotDict

from .filter import Filter


class Created(Filter):

    """
    Matches files by created date

    :param int years:
        specify number of years

    :param int months:
        specify number of months

    :param float weeks:
        specify number of weeks

    :param float days:
        specify number of days

    :param float hours:
        specify number of hours

    :param float minutes:
        specify number of minutes

    :param float seconds:
        specify number of seconds

    :param str mode:
        either 'older' or 'newer'. 'older' matches all files created before the given
        time, 'newer' matches all files created within the given time.
        (default = 'older')

    :returns:
        - ``{created.year}`` -- the year the file was created
        - ``{created.month}`` -- the month the file was created
        - ``{created.day}`` -- the day the file was created
        - ``{created.hour}`` -- the hour the file was created
        - ``{created.minute}`` -- the minute the file was created
        - ``{created.second}`` -- the second the file was created

    Examples:
        - Show all files on your desktop created at least 10 days ago:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - created:
                      days: 10
                actions:
                  - echo: 'Was created at least 10 days ago'

        - Show all files on your desktop which were created within the last 5 hours:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - created:
                      hours: 5
                      mode: newer
                actions:
                  - echo: 'Was created within the last 5 hours'

        - Sort pdfs by year of creation:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents'
                filters:
                  - extension: pdf
                  - created
                actions:
                  - move: '~/Documents/PDF/{created.year}/'
    """

    def __init__(
        self,
        years=0,
        months=0,
        weeks=0,
        days=0,
        hours=0,
        minutes=0,
        seconds=0,
        mode="older",
    ) -> None:
        self._mode = mode.strip().lower()
        if self._mode not in ("older", "newer"):
            raise ValueError("Unknown option for 'mode': must be 'older' or 'newer'.")
        self.is_older = self._mode == "older"
        self.timedelta = pendulum.duration(
            years=years,
            months=months,
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
        )
        print(bool(self.timedelta))

    def pipeline(self, args: DotDict) -> Optional[Dict[str, pendulum.DateTime]]:
        created_date = self._created(args.path)
        # Pendulum bug: https://github.com/sdispater/pendulum/issues/387
        # in_words() is a workaround: total_seconds() returns 0 if years are given
        if self.timedelta.in_words():
            is_past = (created_date + self.timedelta).is_past()
            match = self.is_older == is_past
        else:
            match = True
        if match:
            return {"created": created_date}
        return None

    def _created(self, path: Path) -> pendulum.DateTime:
        # see https://stackoverflow.com/a/39501288/300783
        stat = path.stat()
        time = 0  # type: SupportsFloat
        if sys.platform.startswith("win"):
            time = stat.st_ctime
        else:
            try:
                time = stat.st_birthtime
            except AttributeError:
                # We're probably on Linux. No easy way to get creation dates here,
                # so we'll settle for when its content was last modified.
                time = stat.st_mtime
        return pendulum.from_timestamp(float(time))

    def __str__(self):
        return "[Created] All files %s than %s" % (
            self._mode,
            self.timedelta.in_words(),
        )



from textwrap import indent
from typing import Any, Mapping, Optional, Callable

from organize.utils import DotDict


class Error(Exception):
    pass


class TemplateAttributeError(Error):
    pass


class Action:
    pre_print_hook = None # type: Optional[Callable]

    def run(self, **kwargs) -> Optional[Mapping[str, Any]]:
        return self.pipeline(DotDict(kwargs))

    def pipeline(self, args: DotDict) -> Optional[Mapping[str, Any]]:
        raise NotImplementedError

    def print(self, msg) -> None:
        """ print a message for the user """
        if callable(self.pre_print_hook):
            self.pre_print_hook()  # pylint: disable=not-callable
        print(indent("- [%s] %s" % (self.__class__.__name__, msg), " " * 4))

    @staticmethod
    def fill_template_tags(msg: str, args) -> str:
        try:
            return msg.format(**args)
        except AttributeError as exc:
            cause = exc.args[0]
            raise TemplateAttributeError(
                'Missing template variable %s for "%s"' % (cause, msg)
            )

    def __str__(self) -> str:
        return self.__class__.__name__

    def __repr__(self) -> str:
        return "<%s>" % str(self)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



import textwrap
from typing import Any, Dict, Optional, Sequence

from .filter import Filter


class Python(Filter):

    r"""
    Use python code to filter files.

    :param str code:
        The python code to execute. The code must contain a ``return`` statement.


    :returns:
        - If your code returns ``False`` or ``None`` the file is filtered out,
          otherwise the file is passed on to the next filters.
        - ``{python}`` contains the returned value. If you return a dictionary (for
          example ``return {"some_key": some_value, "nested": {"k": 2}}``) it will be
          accessible via dot syntax in your actions: ``{python.some_key}``,
          ``{python.nested.k}``.

    Examples:
      - A file name reverser.

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Documents
              filters:
              - extension
              - python: |
                  return {"reversed_name": path.stem[::-1]}
              actions:
              - rename: '{python.reversed_name}.{extension}'

      - A filter for odd student numbers. Assuming the folder ``~/Students`` contains
        the files ``student-01.jpg``, ``student-01.txt``, ``student-02.txt`` and
        ``student-03.txt`` this rule will print
        ``"Odd student numbers: student-01.txt"`` and
        ``"Odd student numbers: student-03.txt"``

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Students/
              filters:
               - python: |
                 return int(path.stem.split('-')[1]) % 2 == 1
               actions:
               - echo: 'Odd student numbers: {path.name}'


      - Advanced usecase. You can access data from previous filters in your python code.
        This can be used to match files and capturing names with a regular expression
        and then renaming the files with the output of your python script.

        .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: files
                filters:
                  - extension: txt
                  - regex: (?P<firstname>\w+)-(?P<lastname>\w+)\..*
                  - python: |
                      emails = {
                          "Betts": "dbetts@mail.de",
                          "Cornish": "acornish@google.com",
                          "Bean": "dbean@aol.com",
                          "Frey": "l-frey@frey.org",
                      }
                      if regex.lastname in emails: # get emails from wherever
                          return {"mail": emails[regex.lastname]}
                actions:
                  - rename: '{python.mail}.txt'

        Result:
         - ``Devonte-Betts.txt`` becomes ``dbetts@mail.de.txt``
         - ``Alaina-Cornish.txt`` becomes ``acornish@google.com.txt``
         - ``Dimitri-Bean.txt`` becomes ``dbean@aol.com.txt``
         - ``Lowri-Frey.txt`` becomes ``l-frey@frey.org.txt``
         - ``Someunknown-User.txt`` remains unchanged because the email is not found

    """

    def __init__(self, code) -> None:
        self.code = textwrap.dedent(code)
        if "return" not in self.code:
            raise ValueError("No return statement found in your code!")

    def usercode(self, *args, **kwargs) -> Optional[Any]:
        pass  # will be overwritten by `create_method`

    def create_method(self, name: str, argnames: Sequence[str], code: str) -> None:
        globals_ = globals().copy()
        globals_["print"] = self.print
        locals_ = locals().copy()
        locals_["self"] = self
        funccode = "def {fnc}__({arg}):\n{cod}\n\nself.{fnc} = {fnc}__\n".format(
            fnc=name,
            arg=", ".join(argnames),
            cod=textwrap.indent(textwrap.dedent(code), " " * 4),
        )
        exec(funccode, globals_, locals_)  # pylint: disable=exec-used

    def pipeline(self, args) -> Optional[Dict[str, Any]]:
        self.create_method(name="usercode", argnames=args.keys(), code=self.code)
        result = self.usercode(**args)  # pylint: disable=assignment-from-no-return
        if result not in (False, None):
            return {"python": result}
        return None



import logging
import subprocess
from typing import Mapping

from .action import Action

logger = logging.getLogger(__name__)


class Shell(Action):

    """
    Executes a shell command

    :param str cmd: The command to execute.

    Example:
      - (macOS) Open all pdfs on your desktop:

        .. code-block:: yaml
          :caption: config.yaml

          rules:
            - folders: '~/Desktop'
              filters:
                - extension: pdf
              actions:
                - shell: 'open "{path}"'
    """

    def __init__(self, cmd: str) -> None:
        self.cmd = cmd

    def pipeline(self, args: Mapping) -> None:
        full_cmd = self.fill_template_tags(self.cmd, args)
        self.print("$ %s" % full_cmd)
        if not args["simulate"]:
            # we use call instead of run to be compatible with python < 3.5
            logger.info('Executing command "%s" in shell.', full_cmd)
            subprocess.call(full_cmd, shell=True)

    def __str__(self) -> str:
        return 'Shell(cmd="%s")' % self.cmd



import os
import logging
from typing import Mapping

from organize.compat import Path

from .action import Action

logger = logging.getLogger(__name__)


class Delete(Action):

    """
    Delete a file from disk.

    Deleted files have no recovery option!
    Using the `Trash` action is strongly advised for most use-cases!

    Example:
        - Delete all JPGs and PNGs on the desktop which are older than one year:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
              - filters:
                  - lastmodified:
                      - days: 365
                  - extension:
                      - png
                      - jpg
              - actions:
                  - delete
    """

    def pipeline(self, args: Mapping):
        path = args["path"]  # type: Path
        simulate = args["simulate"]  # type: bool
        self.print('Delete "%s"' % path)
        if not simulate:
            logger.info("Deleting file %s.", path)
            os.remove(str(path))



from typing import Dict, Optional

import pendulum  # type: ignore
from organize.compat import Path
from organize.utils import DotDict

from .filter import Filter


class LastModified(Filter):

    """
    Matches files by last modified date

    :param int years:
        specify number of years

    :param int months:
        specify number of months

    :param float weeks:
        specify number of weeks

    :param float days:
        specify number of days

    :param float hours:
        specify number of hours

    :param float minutes:
        specify number of minutes

    :param float seconds:
        specify number of seconds

    :param str mode:
        either 'older' or 'newer'. 'older' matches all files last modified
        before the given time, 'newer' matches all files last modified within
        the given time. (default = 'older')

    :returns:
        - ``{lastmodified.year}`` -- the year the file was last modified
        - ``{lastmodified.month}`` -- the month the file was last modified
        - ``{lastmodified.day}`` -- the day the file was last modified
        - ``{lastmodified.hour}`` -- the hour the file was last modified
        - ``{lastmodified.minute}`` -- the minute the file was last modified
        - ``{lastmodified.second}`` -- the second the file was last modified

    Examples:
        - Show all files on your desktop last modified at least 10 days ago:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - lastmodified:
                      days: 10
                actions:
                  - echo: 'Was modified at least 10 days ago'

        - Show all files on your desktop which were modified within the last
          5 hours:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - lastmodified:
                      hours: 5
                      mode: newer
                actions:
                  - echo: 'Was modified within the last 5 hours'

        - Sort pdfs by year of last modification

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents'
                filters:
                  - extension: pdf
                  - LastModified
                actions:
                  - move: '~/Documents/PDF/{lastmodified.year}/'
    """

    def __init__(
        self,
        years=0,
        months=0,
        weeks=0,
        days=0,
        hours=0,
        minutes=0,
        seconds=0,
        mode="older",
    ) -> None:
        self._mode = mode.strip().lower()
        if self._mode not in ("older", "newer"):
            raise ValueError("Unknown option for 'mode': must be 'older' or 'newer'.")
        self.is_older = self._mode == "older"
        self.timedelta = pendulum.duration(
            years=years,
            months=months,
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
        )

    def pipeline(self, args: DotDict) -> Optional[Dict[str, pendulum.DateTime]]:
        file_modified = self._last_modified(args.path)
        # Pendulum bug: https://github.com/sdispater/pendulum/issues/387
        # in_words() is a workaround: total_seconds() returns 0 if years are given
        if self.timedelta.in_words():
            is_past = (file_modified + self.timedelta).is_past()
            match = self.is_older == is_past
        else:
            match = True
        if match:
            return {"lastmodified": file_modified}
        return None

    def _last_modified(self, path: Path) -> pendulum.DateTime:
        return pendulum.from_timestamp(float(path.stat().st_mtime))

    def __str__(self):
        return "[LastModified] All files last modified %s than %s" % (
            self._mode,
            self.timedelta.in_words(),
        )



import logging
import logging.config
import os

import appdirs  # type: ignore
import colorama  # type: ignore
import yaml

from .compat import Path

colorama.init(autoreset=True)

# prepare config and log folders
APP_DIRS = appdirs.AppDirs("organize")

# setting the $ORGANIZE_CONFIG env variable overrides the default config path
if os.getenv("ORGANIZE_CONFIG"):
    CONFIG_PATH = Path(os.getenv("ORGANIZE_CONFIG", "")).resolve()
    CONFIG_DIR = CONFIG_PATH.parent
else:
    CONFIG_DIR = Path(APP_DIRS.user_config_dir)
    CONFIG_PATH = CONFIG_DIR / "config.yaml"

LOG_DIR = Path(APP_DIRS.user_log_dir)
LOG_PATH = LOG_DIR / "organize.log"

for folder in (CONFIG_DIR, LOG_DIR):
    folder.mkdir(parents=True, exist_ok=True)

# create empty config file if it does not exist
if not CONFIG_PATH.exists():
    CONFIG_PATH.touch()

# configure logging
LOGGING_CONFIG = """
version: 1
disable_existing_loggers: false
formatters:
    simple:
        format: '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
handlers:
    console:
        class: logging.StreamHandler
        level: DEBUG
        formatter: simple
        stream: ext://sys.stdout
    file:
        class: logging.handlers.TimedRotatingFileHandler
        level: DEBUG
        filename: {filename}
        formatter: simple
        when: midnight
        backupCount: 30
root:
    level: DEBUG
    handlers: [file]
exifread:
    level: INFO
""".format(
    filename=str(LOG_PATH)
)
logging.config.dictConfig(yaml.safe_load(LOGGING_CONFIG))



from typing import Any, List, Mapping, Union

from organize.compat import Path

from .filter import Filter


class Filename(Filter):

    """
    Match files by filename

    :param str startswith:
        The filename must begin with the given string

    :param str contains:
        The filename must contain the given string

    :param str endswith:
        The filename (without extension) must end with the given string

    :param bool case_sensitive = True:
        By default, the matching is case sensitive. Change this to False to use
        case insensitive matching.

    Examples:
        - Match all files starting with 'Invoice':

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - filename:
                      startswith: Invoice
                actions:
                  - echo: 'This is an invoice'

        - Match all files starting with 'A' end containing the string 'hole'
          (case insensitive)

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - filename:
                      startswith: A
                      contains: hole
                      case_sensitive: false
                actions:
                  - echo: 'Found a match.'

        - Match all files starting with 'A' or 'B' containing '5' or '6' and ending with
          '_end'

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - filename:
                      startswith:
                        - A
                        - B
                      contains:
                        - 5
                        - 6
                      endswith: _end
                      case_sensitive: false
                actions:
                  - echo: 'Found a match.'
    """

    def __init__(
        self, startswith="", contains="", endswith="", case_sensitive=True
    ) -> None:
        self.startswith = self.create_list(startswith, case_sensitive)
        self.contains = self.create_list(contains, case_sensitive)
        self.endswith = self.create_list(endswith, case_sensitive)
        self.case_sensitive = case_sensitive

    def matches(self, path: Path) -> bool:
        filename = path.stem
        if not self.case_sensitive:
            filename = filename.lower()
        return (
            any(x in filename for x in self.contains)
            and any(filename.startswith(x) for x in self.startswith)
            and any(filename.endswith(x) for x in self.endswith)
        )

    def pipeline(self, args: Mapping) -> bool:
        return self.matches(args["path"])

    @staticmethod
    def create_list(x: Union[str, List[Any]], case_sensitive: bool) -> List[str]:
        if isinstance(x, str):
            x = [x]
        x = [str(x) for x in x]
        if not case_sensitive:
            x = [x.lower() for x in x]
        return x



import sys

# in python < 3.6 the pathlib module misses some features so we have to import
# a backported alternative
if sys.version_info < (3, 6):
    from pathlib2 import Path  # type: ignore
else:
    from pathlib import Path



import logging
import os
from typing import Mapping

from organize.compat import Path
from organize.utils import find_unused_filename

from .action import Action
from .trash import Trash

logger = logging.getLogger(__name__)


class Rename(Action):

    """
    Renames a file.

    :param str name:
        The new filename.
        Can be a format string which uses file attributes from a filter.

    :param bool overwrite:
        specifies whether existing files should be overwritten.
        Otherwise it will start enumerating files (append a counter to the
        filename) to resolve naming conflicts. [Default: False]

    :param str counter_separator:
        specifies the separator between filename and the appended counter.
        Only relevant if **overwrite** is disabled. [Default: ``\' \'``]

    Examples:
        - Convert all .PDF file extensions to lowercase (.pdf):

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - extension: PDF
                actions:
                  - rename: "{path.stem}.pdf"

        - Convert **all** file extensions to lowercase:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - Extension
                actions:
                  - rename: "{path.stem}.{extension.lower}"
    """

    def __init__(self, name: str, overwrite=False, counter_separator=" ") -> None:
        if os.path.sep in name:
            ValueError(
                "Rename only takes a filename as argument. To move files between "
                "folders use the Move action."
            )
        self.name = name
        self.overwrite = overwrite
        self.counter_separator = counter_separator

    def pipeline(self, args: Mapping) -> Mapping[str, Path]:
        path = args["path"]  # type: Path
        simulate = args["simulate"]
        expanded_name = self.fill_template_tags(self.name, args)
        new_path = path.parent / expanded_name

        # handle filename collisions
        new_path_exists = new_path.exists()
        new_path_samefile = new_path_exists and new_path.samefile(path)
        if new_path_exists and not new_path_samefile:
            if self.overwrite:
                self.print("File already exists")
                Trash().run(path=new_path, simulate=simulate)
            else:
                new_path = find_unused_filename(
                    path=new_path, separator=self.counter_separator
                )

        # do nothing if the new name is equal to the old name and the file is
        # the same
        if new_path_samefile and new_path == path:
            self.print("Keep name")
        else:
            self.print('New name: "%s"' % new_path.name)
            if not simulate:
                logger.info('Renaming "%s" to "%s".', path, new_path)
                path.rename(new_path)

        return {"path": new_path}

    def __str__(self) -> str:
        return "Rename(name=%s, overwrite=%s, sep=%s)" % (
            self.name,
            self.overwrite,
            self.counter_separator,
        )



import logging
import textwrap
from typing import Any, Mapping, Optional, Iterable

from organize.utils import DotDict

from .action import Action

logger = logging.getLogger(__name__)


class Python(Action):

    r"""
    Execute python code in your config file.

    :param str code: The python code to execute

    Examples:
      - A basic example that shows how to get the current file path and do some
        printing in a for loop. The ``|`` is yaml syntax for defining a string
        literal spanning multiple lines.

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: '~/Desktop'
              actions:
                - python: |
                    print('The path of the current file is %s' % path)
                    for _ in range(5):
                        print('Heyho, its me from the loop')

      - You can access filter data:

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Desktop
              filters:
                - regex: '^(?P<name>.*)\.(?P<extension>.*)$'
              actions:
                - python: |
                    print('Name: %s' % regex.name)
                    print('Extension: %s' % regex.extension)

      - You have access to all the python magic -- do a google search for each
        filename starting with an underscore:

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Desktop
              filters:
                - filename:
                    startswith: '_'
              actions:
                - python: |
                    import webbrowser
                    webbrowser.open('https://www.google.com/search?q=%s' % path.stem)
    """

    def __init__(self, code) -> None:
        self.code = textwrap.dedent(code)

    def usercode(self, *args, **kwargs) -> Optional[Any]:
        pass  # will be overwritten by `create_method`

    def create_method(self, name: str, argnames: Iterable[str], code: str) -> None:
        globals_ = globals().copy()
        globals_["print"] = self.print
        locals_ = locals().copy()
        funccode = "def {fnc}__({arg}):\n{cod}\n\nself.{fnc} = {fnc}__\n".format(
            fnc=name,
            arg=", ".join(argnames),
            cod=textwrap.indent(textwrap.dedent(code), " " * 4),
        )
        exec(funccode, globals_, locals_)  # pylint: disable=exec-used

    def pipeline(self, args: DotDict) -> Optional[Mapping[str, Any]]:
        simulate = args.simulate
        if simulate:
            self.print("Code not run in simulation. (Args: %s)" % args)
            return None

        logger.info('Executing python:\n"""\n%s\n""", args=%s', self.code, args)
        self.create_method(name="usercode", argnames=args.keys(), code=self.code)
        self.print("Running python script.")

        result = self.usercode(**args)  # pylint: disable=assignment-from-no-return
        return result



import re
from typing import Any, Dict, Mapping, Optional

from organize.compat import Path

from .filter import Filter


class Regex(Filter):

    r"""
    Matches filenames with the given regular expression

    :param str expr:
        The regular expression to be matched.

    Any named groups in your regular expression will be returned like this:

    :returns:
        - ``{regex.yourgroupname}`` -- The text matched with the named group
          ``(?P<yourgroupname>)``

    Examples:
        - Match an invoice with a regular expression:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - regex: '^RG(\d{12})-sig\.pdf$'
                actions:
                  - move: '~/Documents/Invoices/1und1/'

        - Match and extract data from filenames with regex named groups:
          This is just like the previous example but we rename the invoice using
          the invoice number extracted via the regular expression and the named
          group ``the_number``.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - regex: '^RG(?P<the_number>\d{12})-sig\.pdf$'
                actions:
                  - move: ~/Documents/Invoices/1und1/{regex.the_number}.pdf
    """

    def __init__(self, expr) -> None:
        self.expr = re.compile(expr, flags=re.UNICODE)

    def matches(self, path: Path) -> Any:
        return self.expr.search(path.name)

    def pipeline(self, args: Mapping) -> Optional[Dict[str, Dict]]:
        match = self.matches(args["path"])
        if match:
            result = match.groupdict()
            return {"regex": result}
        return None



"""
organize -- The file management automation tool.

Usage:
    organize sim [--config-file=<path>]
    organize run [--config-file=<path>]
    organize config [--open-folder | --path | --debug] [--config-file=<path>]
    organize list
    organize --help
    organize --version

Arguments:
    sim             Simulate a run. Does not touch your files.
    run             Organizes your files according to your rules.
    config          Open the configuration file in $EDITOR.
    list            List available filters and actions.
    --version       Show program version and exit.
    -h, --help      Show this screen and exit.

Options:
    -o, --open-folder  Open the folder containing the configuration files.
    -p, --path         Show the path to the configuration file.
    -d, --debug        Debug your configuration file.

Full documentation: https://organize.readthedocs.io
"""
import logging
import os
import sys
from typing import Union

from colorama import Fore, Style  # type: ignore
from docopt import docopt  # type: ignore

from . import CONFIG_DIR, CONFIG_PATH, LOG_PATH
from .__version__ import __version__
from .compat import Path
from .config import Config
from .core import execute_rules
from .utils import flatten, fullpath

logger = logging.getLogger("organize")


def main(argv=None):
    """ entry point for the command line interface """
    args = docopt(__doc__, argv=argv, version=__version__, help=True)

    # override default config file path
    if args["--config-file"]:
        expanded_path = os.path.expandvars(args["--config-file"])
        config_path = Path(expanded_path).expanduser().resolve()
        config_dir = config_path.parent
    else:
        config_dir = CONFIG_DIR
        config_path = CONFIG_PATH

    # > organize config
    if args["config"]:
        if args["--open-folder"]:
            open_in_filemanager(config_dir)
        elif args["--path"]:
            print(str(config_path))
        elif args["--debug"]:
            config_debug(config_path)
        else:
            config_edit(config_path)

    # > organize list
    elif args["list"]:
        list_actions_and_filters()

    # > organize sim / run
    else:
        try:
            config = Config.from_file(config_path)
            execute_rules(config.rules, simulate=args["sim"])
        except Config.Error as e:
            logger.exception(e)
            print_error(e)
            print("Try 'organize config --debug' for easier debugging.")
            print("Full traceback at: %s" % LOG_PATH)
            sys.exit(1)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            print_error(e)
            print("Full traceback at: %s" % LOG_PATH)
            sys.exit(1)


def config_edit(config_path: Path) -> None:
    """ open the config file in $EDITOR or default text editor """
    # attention: the env variable might contain command line arguments.
    # https://github.com/tfeldmann/organize/issues/24
    editor = os.getenv("EDITOR")
    if editor:
        os.system('%s "%s"' % (editor, config_path))
    else:
        open_in_filemanager(config_path)


def open_in_filemanager(path: Path) -> None:
    """ opens the given path in file manager, using the default application """
    import webbrowser  # pylint: disable=import-outside-toplevel

    webbrowser.open(path.as_uri())


def config_debug(config_path: Path) -> None:
    """ prints the config with resolved yaml aliases, checks rules syntax and checks
        whether the given folders exist
    """
    print(str(config_path))
    haserr = False
    # check config syntax
    try:
        print(Style.BRIGHT + "Your configuration as seen by the parser:")
        config = Config.from_file(config_path)
        if not config.config:
            print_error("Config file is empty")
            return
        print(config.yaml())
        rules = config.rules
        print("Config file syntax seems fine!")
    except Config.Error as e:
        haserr = True
        print_error(e)
    else:
        # check whether all folders exists:
        allfolders = set(flatten([rule.folders for rule in rules]))
        for f in allfolders:
            if not fullpath(f).exists():
                haserr = True
                print(Fore.YELLOW + 'Warning: "%s" does not exist!' % f)

    if not haserr:
        print(Fore.GREEN + Style.BRIGHT + "No config problems found.")


def list_actions_and_filters() -> None:
    """ Prints a list of available actions and filters """
    import inspect  # pylint: disable=import-outside-toplevel
    from organize import filters, actions  # pylint: disable=import-outside-toplevel

    print(Style.BRIGHT + "Filters:")
    for name, _ in inspect.getmembers(filters, inspect.isclass):
        print("  " + name)
    print()
    print(Style.BRIGHT + "Actions:")
    for name, _ in inspect.getmembers(actions, inspect.isclass):
        print("  " + name)


def print_error(e: Union[Exception, str]) -> None:
    print(Style.BRIGHT + Fore.RED + "ERROR:" + Style.RESET_ALL + " %s" % e)



from .copy import Copy
from .delete import Delete
from .echo import Echo
from .move import Move
from .python import Python
from .rename import Rename
from .shell import Shell
from .trash import Trash



import logging
from typing import Mapping

from organize.compat import Path

from .action import Action

logger = logging.getLogger(__name__)


class Trash(Action):

    """
    Move a file into the trash.

    Example:
        - Move all JPGs and PNGs on the desktop which are older than one year
          into the trash:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
              - filters:
                  - lastmodified:
                      - days: 365
                  - extension:
                      - png
                      - jpg
              - actions:
                  - trash
    """

    def pipeline(self, args: Mapping):
        path = args["path"]  # type: Path
        simulate = args["simulate"]  # type: bool
        from send2trash import send2trash  # type: ignore

        self.print('Trash "%s"' % path)
        if not simulate:
            logger.info("Moving file %s into trash.", path)
            send2trash(str(path))



from typing import Dict, Optional, Union

from organize.compat import Path
from organize.utils import DotDict, flatten

from .filter import Filter


class ExtensionResult:
    def __init__(self, ext):
        self.ext = ext[1:] if ext.startswith(".") else ext

    @property
    def lower(self):
        return self.ext.lower()

    @property
    def upper(self):
        return self.ext.upper()

    def __str__(self):
        return self.ext


class Extension(Filter):

    """
    Filter by file extension

    :param extensions:
        The file extensions to match (does not need to start with a colon).

    :returns:
        - ``{extension}`` -- the original file extension (without colon)
        - ``{extension.lower}`` -- the file extension in lowercase
        - ``{extension.upper}`` -- the file extension in UPPERCASE

    Examples:
        - Match a single file extension:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - extension: png
                actions:
                  - echo: 'Found PNG file: {path}'

        - Match multiple file extensions:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - extension:
                    - .jpg
                    - jpeg
                actions:
                  - echo: 'Found JPG file: {path}'

        - Make all file extensions lowercase:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - Extension
                actions:
                  - rename: '{path.stem}.{extension.lower}'

        - Using extension lists:

          .. code-block:: yaml
            :caption: config.yaml

            img_ext: &img
              - png
              - jpg
              - tiff

            audio_ext: &audio
              - mp3
              - wav
              - ogg

            rules:
              - folders: '~/Desktop'
                filters:
                  - extension:
                    - *img
                    - *audio
                actions:
                  - echo: 'Found media file: {path}'
    """

    def __init__(self, *extensions) -> None:
        self.extensions = list(map(self.normalize_extension, flatten(list(extensions))))

    @staticmethod
    def normalize_extension(ext: str) -> str:
        """ strip colon and convert to lowercase """
        if ext.startswith("."):
            return ext[1:].lower()
        else:
            return ext.lower()

    def matches(self, path: Path) -> Union[bool, str]:
        if not self.extensions:
            return True
        if not path.suffix:
            return False
        return self.normalize_extension(path.suffix) in self.extensions

    def pipeline(self, args: DotDict) -> Optional[Dict[str, ExtensionResult]]:
        if self.matches(args.path):
            result = ExtensionResult(args.path.suffix)
            return {"extension": result}
        return None

    def __str__(self):
        return "Extension(%s)" % ", ".join(self.extensions)



import os
import re
from collections.abc import Mapping
from copy import deepcopy
from typing import Any, Sequence, Tuple, Union, List, Hashable

from .compat import Path

WILDCARD_REGEX = re.compile(r"(?<!\\)[\*\?\[]+")


def splitglob(globstr: str) -> Tuple[Path, str]:
    """ split a string with wildcards into a base folder and globstring """
    path = fullpath(globstr.strip())
    parts = path.parts
    for i, part in enumerate(parts):
        if WILDCARD_REGEX.search(part):
            return (Path(*parts[:i]), str(Path(*parts[i:])))
    return (path, "")


def fullpath(path: Union[str, Path]) -> Path:
    """ Expand '~' and resolve the given path. Path can be a string or a Path obj. """
    return Path(os.path.expandvars(str(path))).expanduser().resolve(strict=False)


def flatten(arr: List[Any]) -> List[Any]:
    if arr == []:
        return []
    if not isinstance(arr, list):
        return [arr]
    return flatten(arr[0]) + flatten(arr[1:])


def flattened_string_list(x, case_sensitive=True) -> Sequence[str]:
    x = [str(x) for x in flatten(x)]
    if not case_sensitive:
        x = [x.lower() for x in x]
    return x


def first_key(dic: Mapping) -> Hashable:
    return list(dic.keys())[0]


class DotDict(dict):
    """
    Quick and dirty implementation of a dot-able dict, which allows access and
    assignment via object properties rather than dict indexing.
    Keys are case insensitive.
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        # we could just call super(DotDict, self).__init__(*args, **kwargs)
        # but that won't get us nested dotdict objects
        od = dict(*args, **kwargs)
        for key, val in od.items():
            if isinstance(val, Mapping):
                value = DotDict(val)
            else:
                value = val
            self[self.normkey(key)] = value

    @staticmethod
    def normkey(key):
        if isinstance(key, str):
            return key.lower()
        else:
            return key

    def __delattr__(self, key):
        try:
            del self[self.normkey(key)]
        except KeyError as ex:
            raise AttributeError("No attribute called: %s" % key) from ex

    def __getattr__(self, key):
        try:
            return self[self.normkey(key)]
        except KeyError as ex:
            raise AttributeError("No attribute called: %s" % key) from ex

    def __setattr__(self, key, value) -> None:
        self[self.normkey(key)] = value

    def update(self, other):
        """ recursively update the dotdict instance with another dicts items """
        for key, val in other.items():
            normkey = self.normkey(key)
            if isinstance(val, Mapping):
                if isinstance(self.get(normkey), dict):
                    self[normkey].update(val)
                else:
                    self[normkey] = __class__(val)
            else:
                self[normkey] = val

    def merge(self, other) -> Mapping:
        """ recursively merge values from another dict and return a new instance """
        new_dct = deepcopy(self)
        new_dct.update(other)
        return new_dct


def increment_filename_version(path: Path, separator=" ") -> Path:
    stem = path.stem
    try:
        # try to find any existing counter
        splitstem = stem.split(separator)  # raises ValueError on missing sep
        if len(splitstem) < 2:
            raise ValueError()
        counter = int(splitstem[-1])
        stem = separator.join(splitstem[:-1])
    except (ValueError, IndexError):
        # not found, we start with 1
        counter = 1
    return path.with_name(
        "{stem}{sep}{cnt}{suffix}".format(
            stem=stem, sep=separator, cnt=(counter + 1), suffix=path.suffix
        )
    )


def find_unused_filename(path: Path, separator=" ") -> Path:
    """
    We assume the given path already exists. This function adds a counter to the
    filename until we find a unused filename.
    """
    # TODO: Check whether the assumption can be eliminated for cleaner code.
    # TODO: Optimization: The counter only needs to be parsed once.
    tmp = path
    while True:
        tmp = increment_filename_version(tmp, separator=separator)
        if not tmp.exists():
            return tmp


def dict_merge(dct, merge_dct, add_keys=True):
    """ Recursive dict merge.

    Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    This version will return a copy of the dictionary and leave the original
    arguments untouched.

    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dict`` but not ``dct`` should be included in the
    new dict.

    Args:
        dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys

    Returns:
        dict: updated dict

    Taken from comment thread: https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
    """
    dct = deepcopy(dct)
    if not add_keys:
        merge_dct = {k: merge_dct[k] for k in set(dct).intersection(set(merge_dct))}

    for k, v in merge_dct.items():
        if isinstance(dct.get(k), dict) and isinstance(v, Mapping):
            dct[k] = dict_merge(dct[k], v, add_keys=add_keys)
        else:
            dct[k] = v

    return dct



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['organize', 'organize.actions', 'organize.filters']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.1,<6.0',
 'Send2Trash>=1.5,<2.0',
 'appdirs>=1.4,<2.0',
 'colorama>=0.4.1,<0.5.0',
 'docopt>=0.6.2,<0.7.0',
 'exifread>=2.1,<3.0',
 'pendulum>=2.0.5,<3.0.0',
 'textract>=1.6.3,<2.0.0']

extras_require = \
{':python_version < "3.4"': ['pathlib2==2.3.0'],
 ':python_version < "3.5"': ['typing>=3.6,<4.0', 'pathlib2>=2.3.3,<3.0.0']}

entry_points = \
{'console_scripts': ['organize = organize.cli:main']}

setup_kwargs = {
    'name': 'organize-tool',
    'version': '1.8.1',
    'description': 'The file management automation tool',
    'long_description': '<p align="center">\n <img width="623" height="168" src="https://github.com/tfeldmann/organize/raw/master/docs/images/organize.svg?sanitize=true" alt="organize logo">\n</p>\n\n<div align="center">\n\n  [![GitHub Issues](https://travis-ci.org/tfeldmann/organize.svg?branch=master)](https://travis-ci.org/tfeldmann/organize)\n  [![Documentation Status](https://readthedocs.org/projects/organize/badge/?version=latest)](https://organize.readthedocs.io/en/latest/?badge=latest)\n  [![License](https://img.shields.io/badge/license-MIT-blue.svg)](/LICENSE)\n</div>\n\n---\n\n<p align="center"> <b>organize</b> - The file management automation tool\n<br>\n<a href="https://organize.readthedocs.io/" target="_blank">Full documentation at Read the docs</a>\n</p>\n\n- [About](#about)\n- [Getting started](#getting-started)\n  - [Installation](#installation)\n  - [Creating your first rule](#creating-your-first-rule)\n- [Example rules](#example-rules)\n- [Advanced usage](#advanced-usage)\n- [Command line interface](#command-line-interface)\n\n## About\nYour desktop is a mess? You cannot find anything in your downloads and\ndocuments? Sorting and renaming all these files by hand is too tedious?\nTime to automate it once and benefit from it forever.\n\n**organize** is a command line, open-source alternative to apps like Hazel (macOS)\nor File Juggler (Windows).\n\n\n## Getting started\n### Installation\nPython 3.5+ is needed. Install it via your package manager or from [python.org](https://python.org).\n\nInstallation is done via pip. Note that the package name is `organize-tool`:\n```bash\npip3 install -U organize-tool\n``` \nThis command can also be used to update to the newest version. Now you can run `organize --help` to check if the installation was successful.\n\n### Creating your first rule\nIn your shell, **run ``organize config``** to edit the configuration:\n\n```yaml\nrules:\n  - folders: ~/Downloads\n    subfolders: true\n    filters:\n      - extension: pdf\n    actions:\n      - echo: "Found PDF!"\n```\n\n> If you have problems editing the configuration you can run ``organize config --open-folder`` to reveal the configuration folder in your file manager. You can then edit the `config.yaml` in your favourite editor.\n> \n> Alternatively you can run ``organize config --path`` to see the full path to\nyour ``config.yaml``)\n\n\n**Save your config file and run `organize run`.**\n\nYou will see a list of all `.pdf` files you have in your downloads folder (+ subfolders). For now we only show the text `Found PDF!` for each file, but this will change soon...\n(If it shows ``Nothing to do`` you simply don\'t have any pdfs in your downloads folder).\n\nRun ``organize config`` again and add a `copy`-action to your rule:\n```yaml\n    actions:\n      - echo: "Found PDF!"\n      - move: ~/Documents/PDFs/\n```\n\n**Now run `organize sim` to see what would happen without touching your files**. You will see that your pdf-files would be moved over to your `Documents/PDFs` folder.\n\nCongratulations, you just automated your first task. You can now run `organize run` whenever you like and all your pdfs are a bit more organized. It\'s that easy.\n\n> There is so much more. You want to rename / copy files, run custom shell- or python scripts, match filenames with regular expressions or use placeholder variables? organize has you covered. Have a look at the advanced usage example below!\n\n\n## Example rules\nHere are some examples of simple organization and cleanup rules. Modify to your needs!\n\nMove all invoices, orders or purchase documents into your documents folder:\n```yaml\nrules:\n  # sort my invoices and receipts\n  - folders: ~/Downloads\n    subfolders: true\n    filters:\n      - extension: pdf\n      - filename:\n          contains:\n            - Invoice\n            - Order\n            - Purchase\n          case_sensitive: false\n    actions:\n      - move: ~/Documents/Shopping/\n```\n\nMove incomplete downloads older than 30 days into the trash:\n```yaml\nrules:\n  # move incomplete downloads older > 30 days into the trash\n  - folders: ~/Downloads\n    filters:\n      - extension:\n          - download\n          - crdownload\n          - part\n      - lastmodified:\n          days: 30\n          mode: older\n    actions:\n      - trash\n```\n\nDelete empty files from downloads and desktop:\n```yaml\nrules:\n  # delete empty files from downloads and desktop\n  - folders: \n      - ~/Downloads\n      - ~/Desktop\n    filters:\n      - filesize: 0\n    actions:\n      - trash\n```\n\nMove screenshots into a "Screenshots" folder on your desktop:\n```yaml\nrules:\n  # move screenshots into "Screenshots" folder\n  - folders: ~/Desktop\n    filters:\n      - filename:\n          startswith: \'Screen Shot\'\n    actions:\n      - move: ~/Desktop/Screenshots/\n```\n\nOrganize your font downloads:\n```yaml\nrules:\n  # organize your font files but keep the folder structure:\n  #   "~/Downloads/favourites/helvetica/helvetica-bold.ttf"\n  #     is moved to\n  #   "~/Documents/FONTS/favourites/helvetica/helvetica-bold.ttf"\n  - folders: ~/Downloads/**/*.ttf\n    actions:\n      - Move: \'~/Documents/FONTS/{relative_path}\'\n```\n\nYou\'ll find many more examples in the <a href="https://organize.readthedocs.io/" target="_blank">full documentation</a>.\n\n\n## Advanced usage\nThis example shows some advanced features like placeholder variables, pluggable\nactions, recursion through subfolders and glob syntax:\n\n```yaml\nrules:\n  - folders: ~/Documents/**/*\n    filters:\n      - extension:\n          - pdf\n          - docx\n      - created\n    actions:\n      - move: \'~/Documents/{extension.upper}/{created.year}{created.month:02}/\'\n      - shell: \'open "{path}"\'\n```\n\nGiven we have two files in our ``~/Documents`` folder (or any of its subfolders)\nnamed ``script.docx`` from january 2018 and ``demo.pdf`` from december 2016 this will\nhappen:\n\n- ``script.docx`` will be moved to ``~/Documents/DOCX/2018-01/script.docx``\n- ``demo.pdf`` will be moved to ``~/Documents/PDF/2016-12/demo.pdf``\n- The files will be opened (``open`` command in macOS) *from their new location*.\n- Note the format syntax for `{created.month}` to make sure the month is prepended with a zero.\n\n\n## Command line interface\n```\nThe file management automation tool.\n\nUsage:\n    organize sim [--config-file=<path>]\n    organize run [--config-file=<path>]\n    organize config [--open-folder | --path | --debug] [--config-file=<path>]\n    organize list\n    organize --help\n    organize --version\n\nArguments:\n    sim             Simulate a run. Does not touch your files.\n    run             Organizes your files according to your rules.\n    config          Open the configuration file in $EDITOR.\n    list            List available filters and actions.\n    --version       Show program version and exit.\n    -h, --help      Show this screen and exit.\n\nOptions:\n    -o, --open-folder  Open the folder containing the configuration files.\n    -p, --path         Show the path to the configuration file.\n    -d, --debug        Debug your configuration file.\n\nFull documentation: https://organize.readthedocs.io\n```\n',
    'author': 'Thomas Feldmann',
    'author_email': 'mail@tfeldmann.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tfeldmann/organize',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from textwrap import indent
from typing import Any, Callable, Dict, Optional, Union

from organize.utils import DotDict

FilterResult = Union[Dict[str, Any], bool, None]


class Filter:
    pre_print_hook = None  # type: Optional[Callable]

    def run(self, **kwargs: Dict) -> FilterResult:
        return self.pipeline(DotDict(kwargs))

    def pipeline(self, args: DotDict) -> FilterResult:
        raise NotImplementedError

    def print(self, msg: str) -> None:
        """ print a message for the user """
        if callable(self.pre_print_hook):
            self.pre_print_hook()  # pylint: disable=not-callable
        print(indent("- (%s) %s" % (self.__class__.__name__, msg), " " * 4))

    def __str__(self) -> str:
        """ Return filter name and properties """
        return self.__class__.__name__

    def __repr__(self) -> str:
        return "<%s>" % str(self)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



from .created import Created
from .exif import Exif
from .extension import Extension
from .file_content import FileContent
from .filename import Filename
from .filesize import FileSize
from .last_modified import LastModified
from .python import Python
from .regex import Regex



import collections
from typing import Any, DefaultDict, Dict, Mapping, Optional, Union

import exifread  # type: ignore

from organize.compat import Path

from .filter import Filter

ExifDict = Mapping[str, Union[str, Mapping[str, str]]]


class Exif(Filter):

    """
    Filter by image EXIF data

    The `exif` filter can be used as a filter as well as a way to get exif information
    into your actions.

    :returns:
        ``{exif}`` -- a dict of all the collected exif inforamtion available in the
        file. Typically it consists of the following tags (if present in the file):

        - ``{exif.image}`` -- information related to the main image
        - ``{exif.exif}`` -- Exif information
        - ``{exif.gps}`` -- GPS information
        - ``{exif.interoperability}`` -- Interoperability information

    Examples:
        - Show available EXIF data of your pictures:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Pictures
                subfolders: true
                filters:
                  - exif
                actions:
                  - echo: "{exif}"

        - Copy all images which contain GPS information while keeping subfolder
          structure:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Pictures
              subfolders: true
              filters:
                - exif:
                    gps.gpsdate
              actions:
                - copy: ~/Pictures/with_gps/{relative_path}/

        - Filter by camera manufacturer:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Pictures
                subfolders: true
                filters:
                  - exif:
                      image.model: Nikon D3200
                actions:
                  - move: '~/Pictures/My old Nikon/'

        - Sort images by camera manufacturer. This will create folders for each camera
          model (for example "Nikon D3200", "iPhone 6s", "iPhone 5s", "DMC-GX80") and
          move the pictures accordingly:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Pictures
                subfolders: true
                filters:
                  - extension: jpg
                  - exif:
                      image.model
                actions:
                  - move: '~/Pictures/{exif.image.model}/'
    """

    def __init__(self, *required_tags: str, **tag_filters: str) -> None:
        self.args = required_tags  # expected exif keys
        self.kwargs = tag_filters  # exif keys with expected values

    def category_dict(self, tags: Mapping[str, str]) -> ExifDict:
        result = collections.defaultdict(dict)  # type: DefaultDict[str, Dict[str, str]]
        for key, value in tags.items():
            if " " in key:
                category, field = key.split(" ", maxsplit=1)
                result[category][field] = value
            else:
                result[key] = value  # type: ignore
        return result

    def matches(self, path: Path) -> Union[bool, ExifDict]:
        # NOTE: This should return Union[Literal[False], ExifDict] but Literal is only
        # available in Python>=3.8.
        with path.open("rb") as f:
            exiftags = exifread.process_file(f, details=False)  # type: Dict
        if not exiftags:
            return False

        tags = {k.lower(): v.printable for k, v in exiftags.items()}

        # no match if expected tag is not found
        normkey = lambda k: k.replace(".", " ").lower()
        for key in self.args:
            if normkey(key) not in tags:
                return False
        # no match if tag has not expected value
        for key, value in self.kwargs.items():
            key = normkey(key)
            if not (key in tags and tags[key].lower() == value.lower()):
                return False
        return self.category_dict(tags)

    def pipeline(self, args: Mapping[str, Any]) -> Optional[Dict[str, ExifDict]]:
        tags = self.matches(args["path"])
        if isinstance(tags, dict):
            return {"exif": tags}
        return None

    def __str__(self) -> str:
        return "EXIF(%s)" % ", ".join("%s=%s" % (k, v) for k, v in self.kwargs.items())
